#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution


# docstring for argument parsing using docopts
"""
PyPEF - Pythonic Protein Engineering Framework

written by Niklas Siedhoff and Alexander-Maurice Illig.

Modeling options
----------------
    I. Pure ML modeling
    -------------------
        PyPEF provides three encoding options for training machine learning models, i.e.
        regression models trained by supervised learning:

            1. DCA: Direct coupling analysis (DCA) based on evolutionary couplings (input:
                       coupling parameter file generated by the C framework plmc) or
                       generating parameters using TensorFlow-based GREMLIN (input: MSA).

            2. AAidx: Based on AAindex descriptors (566 amino acid descriptor files
                       taken from the AAindex database).

            3. OneHot: One-hot encoding representing the occurrence of an
                       amino acid at a sequence position as a single 1 and 19 0's.

        Any encoding technique enables pure ML-based modeling, see
        https://doi.org/10.1021/acs.jcim.1c00099
        and DCA-based sequence encoding enables a hybrid modeling approach, see
        https://doi.org/10.1101/2022.06.07.495081

        If an MSA can be constructed for the target sequence, e.g. using Jackhmmer,
        encoding option 1 will likely outperform encoding option 2.
        However, encoding option 2 provides a static encoding technique that is
        independent from the evolutionary history of a target sequence and
        without the need for MSA construction.
        Here, the AAidx encodings for modeling are compared, i.e. validated, with respect
        to their performance on the test set (comparable to an hyperparameter search
        for finding the best static encoding set for model inference).
        Further, one-hot encoding (encoding option 3) provides a simple but fast and often
        well-performing encoding option that will likely outperform the AAindex-based
        technique for model generalization.

    II. Hybrid modeling
    -------------------
        Constructing a hybrid model that combines pure statistical DCA-based prediction (a
        variant's relative 'evolutionary energy' to the wild type) and DCA-encoding based
        training of a ML model similar to pure ML modeling option I.1.
        Based on features generated from the direct coupling analysis (.params file output
        using the plmc framework or provided MSA and running GREMLIN).
        Individual model contributions are optimization only based on Spearman's correlation
        coefficient and thus, only variant fitness ranks are to be considered for evaluating
        model performance, not the exact predicted fitness value. For regression, up to now
        only L2-regularized linear regression (Ridge regression) is provided as modeling option.


Running example of training, testing, and using a pure ML model for prediction
------------------------------------------------------------------------------
Exemplary running of PyPEF for training a pure ML model using encoding option 2
based on features generated from the AAIndex database (566 amino acid descriptor
indices taken from the AAIndex database).
 1. Create files for training and testing from variant-fitness CSV data:
       pypef mklsts -i variant_and_fitness.csv -w wt_sequence.fasta
 2. Train and validate models:
        pypef ml -e onehot -l LS.fasta -t TS.fasta --regressor pls
 3. Plot the test set entries against test set predictions (creates PNG figure, MODEL is
    the chosen AAindex, the ML-DCA model, or here the ONEHOT model):
        pypef ml -e onehot -m ONEHOT -t TS.fasta
 4. Create files for prediction:
    - Single file:
        pypef mkps -w wt_sequence.fasta -i variant_fitness.csv
    - Recombinant/diverse prediction files:
        pypef mkps -w wt_sequence.fasta -i variant_fitness.csv
            [--drecomb] [--trecomb] [--qarecomb] [--qirecomb]
            [--ddiverse] [--tdiverse] [--qdiverse]
 5. Predict (unknown/new) variants:
    - Single file:
        pypef ml -e aaidx -m MODEL -p Prediction_Set.fasta
    - Recombinant/diverse prediction files in created prediction set folders:
        pypef ml -e aaidx -m MODEL --pmult [--drecomb] [...] [--qdiverse]
    - Directed evolution – for performing and plotting in silico evolution trajectories:
        pypef ml -e aaidx directevo -m MODEL [...]
Note: The commands for hybrid modeling are very similar to the commands for pure ML modeling,
see pypef -h for possible commands.

For generating DCA parameters using GREMLIN, you have to provide an MSA in FASTA or A2M format:
pypef param_inference --msa MSA_FILE --wt WT_FASTA [--opt_iter 100]


Helpful commands for data conversion
-----------------------------------------------
Creation of learning and test sets – splitting CSV variant-fitness data:
        pypef mklsts --wt WT_FASTA --input CSV_FILE
            [--drop THRESHOLD] [--numrnd NUMBER]

Creation of prediction sets from CSV data (using single-substituted variant data):
        pypef mkps --wt WT_FASTA --input CSV_FILE
            [--drop THRESHOLD] [--drecomb] [--trecomb] [--qarecomb] [--qirecomb]
            [--ddiverse] [--tdiverse] [--qdiverse]

Encoding a CSV file (for further performance studies such as "low N" or
"mutational extrapolation" engineering tasks:
        pypef encode --input CSV_FILE --encoding ENCODING_TECHNIQUE --wt WT_FASTA
            [--params PARAM_FILE] [--y_wt WT_FITNESS] [--model MODEL] [--nofft]
            [--threads THREADS] [--sep CSV_COLUMN_SEPARATOR] [--fitness_key FITNESS_KEY]

Converting a STO alignment file to A2M format:
        pypef sto2a2m --sto STO_MSA_FILE
            [--inter_gap INTER_GAP] [--intra_gap INTRA_GAP]


Usage:
    pypef mklsts --wt WT_FASTA --input CSV_FILE
        [--drop THRESHOLD] [--sep CSV_COLUMN_SEPARATOR] [--mutation_sep MUTATION_SEPARATOR] [--numrnd NUMBER]
    pypef mkps --wt WT_FASTA [--input CSV_FILE]
        [--drop THRESHOLD] [--ssm] [--drecomb] [--trecomb] [--qarecomb] [--qirecomb]
        [--ddiverse] [--tdiverse] [--qdiverse]
    pypef param_inference
        [--msa MSA_FILE] [--params PARAM_FILE]
        [--wt WT_FASTA] [--opt_iter N_ITER]
    pypef save_msa_info --msa MSA_FILE --wt WT_FASTA
        [--opt_iter N_ITER]
    pypef encode --input CSV_FILE --encoding ENCODING_TECHNIQUE --wt WT_FASTA
        [--params PARAM_FILE] [--y_wt WT_FITNESS] [--model MODEL] [--nofft]
        [--threads THREADS]
        [--sep CSV_COLUMN_SEPARATOR] [--fitness_key FITNESS_KEY]
    pypef reformat_csv --input CSV_FILE
        [--sep CSV_COLUMN_SEPARATOR] [--mutation_sep MUTATION_SEPARATOR] [--fitness_key FITNESS_KEY]
    pypef shift_pos --input CSV_FILE --offset OFFSET
        [--sep CSV_COLUMN_SEPARATOR] [--mutation_sep MUTATION_SEPARATOR] [--fitness_key FITNESS_KEY]
    pypef sto2a2m --sto STO_MSA_FILE [--inter_gap INTER_GAP] [--intra_gap INTRA_GAP]
    pypef hybrid --ts TEST_SET
        [--model MODEL] [--params PARAM_FILE]
        [--ls LEARNING_SET] [--label] [--threads THREADS]
    pypef hybrid --model MODEL --params PARAM_FILE
        [--ts TEST_SET] [--label]
        [--ps PREDICTION_SET] [--pmult] [--drecomb] [--trecomb] [--qarecomb] [--qirecomb]
                                        [--ddiverse] [--tdiverse] [--qdiverse] [--negative]
        [--threads THREADS]
    pypef hybrid directevo --wt WT_FASTA --params PARAM_FILE
        [--model MODEL]
        [--input CSV_FILE] [--y_wt WT_FITNESS] [--numiter NUM_ITER]
        [--numtraj NUM_TRAJ] [--temp TEMPERATURE]
        [--negative] [--usecsv] [--csvaa] [--drop THRESHOLD]
    pypef hybrid train_and_save --input CSV_FILE --params PARAM_FILE --wt WT_FASTA
        [--fit_size REL_LEARN_FIT_SIZE] [--test_size REL_TEST_SIZE]
        [--threads THREADS] [--sep CSV_COLUMN_SEPARATOR]
        [--fitness_key FITNESS_KEY] [--rnd_state RND_STATE]
    pypef hybrid low_n --input ENCODED_CSV_FILE
    pypef hybrid extrapolation --input ENCODED_CSV_FILE
        [--conc]
    pypef ml --encoding ENCODING_TECHNIQUE --ls LEARNING_SET --ts TEST_SET
        [--save NUMBER] [--regressor TYPE] [--nofft] [--all] [--params PARAM_FILE]
        [--sort METRIC_INT] [--threads THREADS] [--label]
    pypef ml --encoding ENCODING_TECHNIQUE --model MODEL --ts TEST_SET
        [--nofft] [--params PARAM_FILE] [--threads THREADS] [--label]
    pypef ml --show
        [MODELS]
    pypef ml --encoding ENCODING_TECHNIQUE --model MODEL --ps PREDICTION_SET
        [--params PARAM_FILE] [--threads THREADS] [--nofft] [--negative]
    pypef ml --encoding ENCODING_TECHNIQUE --model MODEL --pmult
        [--drecomb] [--trecomb] [--qarecomb] [--qirecomb]
        [--ddiverse] [--tdiverse] [--qdiverse]
        [--regressor TYPE] [--nofft] [--negative] [--params PARAM_FILE] [--threads THREADS]
    pypef ml --encoding ENCODING_TECHNIQUE directevo --model MODEL --wt WT_FASTA
        [--input CSV_FILE] [--y_wt WT_FITNESS] [--numiter NUM_ITER] [--numtraj NUM_TRAJ] [--temp TEMPERATURE]
        [--nofft] [--negative] [--usecsv] [--csvaa] [--drop THRESHOLD] [--params PARAM_FILE]
    pypef ml low_n --input ENCODED_CSV_FILE
        [--regressor TYPE]
    pypef ml extrapolation --input ENCODED_CSV_FILE
        [--regressor TYPE] [--conc]


Options:
  --all                             Finally training on all data [default: False].
  --conc                            Concatenating mutational level variants for predicting variants
                                    from next higher level [default: False].
  --csvaa                           Directed evolution csv amino acid substitutions,
                                    requires flag "--usecsv" [default: False].
  --ddiverse                        Create/predict double natural diverse variants [default: False].
  --drecomb                         Create/predict double recombinants [default: False].
  -d --drop THRESHOLD               Below threshold variants will be discarded from the
                                    data [default: -9E09].
  -e --encoding ENCODING_TECHNIQUE  Sets technique used for encoding sequences for constructing regression models;
                                    choose between 'aaidx' (AAIndex-based encoding), 'onehot' (OneHot-based encoding),
                                    and DCA encoding using Gremlin/plmc (DCA-based encoding) [default: onehot].
  --fitness_key FITNESS_KEY         Label of CSV fitness column. Else uses second column.
  -h --help                         Show this screen [default: False].
  -i --input CSV_FILE               Input data file in .csv format.
  --inter_gap INTER_GAP             Fraction to delete all positions with more than
                                    'inter_gap' * 100 % gaps (columnar trimming) [default: 0.3].
  --intra_gap INTRA_GAP             Fraction to delete all sequences with more than
                                    'intra_gap' * 100 % gaps after being columnar trimmed
                                    (line trimming) [default: 0.5].
  --label                           Label the plot instances [default: False].
  -l --ls LEARNING_SET              Input learning set in .fasta format.
  -m --model MODEL                  Model (pickle file) for plotting of validation or for
                                    performing predictions.
  --msa MSA_FILE                    Multiple sequence alignment (MSA) ins FASTA or A2M format for
                                    inferring DCA parameters.
  --mutation_sep MUTATION_SEP       Mutation separator [default: /].
  --mutation_extrapolation          Mutation extrapolation [default: False].
  --negative                        Set if more negative values define better variants [default: False].
  --nofft                           Raw sequence input, i.e., no FFT for establishing protein spectra
                                    as vector inputs, only implemented as option for AAindex-based
                                    sequence encoding [default: False].
  -n --numrnd NUMBER                Number of randomly created Learning and Validation
                                    datasets [default: 0].
  --numiter NUM_ITER                Number of mutation iterations per evolution trajectory [default: 5].
  --numtraj NUM_TRAJ                Number of trajectories, i.e., evolution pathways [default: 5].
  -o --offset OFFSET                Offset for shifting substitution positions of the input CSV file [default: 0].
  --opt_iter N_ITER                 Number of iterations for GREMLIN-based optimization of local fields
                                    and couplings [default: 100].
  --params PARAM_FILE               Input PLMC couplings parameter file.
  -u --pmult                        Predict for all prediction files in folder for recombinants
                                    or for diverse variants [default: False].
  -p --ps PREDICTION_SET            Prediction set for performing predictions using a trained Model.
  --qdiverse                        Create quadruple natural diverse variants [default: False].
  --qarecomb                        Create/predict quadruple recombinants [default: False].
  --qirecomb                        Create/predict quintuple recombinants [default: False].
  --regressor TYPE                  Type of regression (R.) to use, options: PLS CV R.: pls,
                                    PLS LOOCV R.: pls_loocv, Random Forest CV R.: rf, SVM CV R.: svr,
                                    MLP CV R.: mlp, Ridge CV R.: ridge (or l2),
                                    LassoLars CV R.: lassolars (or l1) [default: pls].
  --rnd_splits RND_SPLITS           Number of random splits for Low N testing [default: 5].
  --rnd_state RND_STATE             Sets the random state for reproduction, only implemented
                                    for hybrid train_and_save [default: 42].
  -s --save NUMBER                  Number of models to be saved as pickle files [default: 5].
  --sep CSV_COLUMN_SEPARATOR        CSV Column separator [default: ;].
  --show                            Show achieved model performances from Model_Results.txt.
  --sort METRIC_INT                 Rank models based on metric {1: R^2, 2: RMSE, 3: NRMSE,
                                    4: Pearson's r, 5: Spearman's rho} [default: 1].
  --ssm                             Create single-saturation mutagenesis prediction set (does not
                                    require CSV input) [default: False].
  --sto STO_MSA_FILE                The input MSA file in STO (Stockholm) format.
  --tdiverse                        Create/predict triple natural diverse variants [default: False].
  --temp TEMPERATURE                "Temperature" of Metropolis-Hastings criterion [default: 0.01]
  --threads THREADS                 Parallel computing of training and validation of models.
                                    Number of threads used in parallel computing, by default
                                    no hyperthreading.
  --fit_size REL_LEARN_FIT_SIZE     Relative size of the train set for initial fitting. The remaining data
                                    for training is used for hyperparameter optimization on train subsets
                                    used for validation, while in sum the total data for training is
                                    training data = train_fit data + train_test(validation) data
                                                  = all data - test data.
                                    The default of 0.66 means that 34 % of the train data is taken for
                                    train_test validation [default: 0.66].
  --test_size REL_TEST_SIZE         Relative size of the test set; if set to 0.0 the trained model
                                    will not be tested [default: 0.2].
  --trecomb                         Create/predict triple recombinants [default: False].
  --usecsv                          Perform directed evolution on single variant csv position
                                    data [default: False].
  -t --ts TEST_SET                  Input validation set in .fasta format.
  --version                         Show version [default: False].
  -w --wt WT_FASTA                  Input wild-type sequence file (in FASTA format).
  --wt_pos WT_POSITION              Row position of encoded wild-type in encoding CSV file (0-indexed) [default: 0].
  -y --y_wt WT_FITNESS              Fitness value (y) of wild-type [default: 1.0].
  encode                            Encoding [default: False].
  hybrid                            Hybrid modeling based on DCA-derived sequence encoding [default: False].
  ml                                Pure machine learning modeling based on encoded sequences [default: False].
  MODELS                            Number of saved models to show [default: 5].
  onehot                            OneHot-based encoding [default: False].
  param_inference                   Inferring DCA params using the GREMLIN approach [default: False].
  reformat_csv                      Reformat input CSV with indicated column and mutation separators to default
                                    CSV style (column separator ';' and mutation separator '/') [default: False.]
  save_msa_info                     Optimize local fields and couplings of MSA based on GREMLIN DCA approach and
                                    save resulting coupling matrix and highly coevolved amino acids.
  shift_pos                         Shift positions of all variant substitutions of the input CSV
                                    file (identical to reformat_csv when setting --offset to 0) [default: False.]
  sto2a2m                           Transform multiple sequence alignment from STO format to
                                    A2M format [default: False].
"""


from os import environ
try:
    environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  # only print TensorFlow errors, set to '0' or comment
except KeyError:                           # lines for seeing TensorFlow infos and warnings
    pass

from sys import argv, version_info
from pypef import __version__
if version_info[0] < 3 or version_info[1] < 9:
    raise SystemError(f"The current version of PyPEF (v {__version__}) "
                      f"requires at least Python 3.9 or higher.")

import time
from datetime import timedelta
import logging

from docopt import docopt
from schema import Schema, SchemaError, Optional, Or, Use

from pypef.ml.ml_run import run_pypef_pure_ml
from pypef.dca.dca_run import run_pypef_hybrid_modeling
from pypef.utils.utils_run import run_pypef_utils

logger = logging.getLogger("pypef")
logger.setLevel(logging.INFO)

ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

formatter = logging.Formatter(
    "%(asctime)s.%(msecs)03d %(levelname)s %(filename)s:%(lineno)d -- %(message)s",
    "%Y-%m-%d %H:%M:%S"
)
ch.setFormatter(formatter)
logger.addHandler(ch)


schema = Schema({
    Optional('--all'): bool,
    Optional('--conc'): bool,
    Optional('--csvaa'): bool,
    Optional('--ddiverse'): bool,
    Optional('--drecomb'): bool,
    Optional('--drop'): Use(float),
    Optional('--encoding'): Use(str),
    Optional('--fitness_key'): Or(None, str),
    Optional('--fit_size'): Use(float),
    Optional('--help'): bool,
    Optional('--input'): Or(None, str),
    Optional('--inter_gap'): Use(float),
    Optional('--intra_gap'): Use(float),
    Optional('--label'): bool,
    Optional('--ls'): Or(None, str),
    Optional('--model'): Or(None, str),
    Optional('--msa'): Or(None, str),
    Optional('--mutation_sep'): Or(None, str),
    Optional('--negative'): bool,
    Optional('--nofft'): bool,
    Optional('--numrnd'): Use(int),
    Optional('--numiter'): Use(int),
    Optional('--numtraj'): Use(int),
    Optional('--offset'): Use(int),
    Optional('--opt_iter'): Use(int),
    Optional('--params'): Or(None, str),
    Optional('--pmult'): bool,
    Optional('--ps'): Or(None, str),
    Optional('--qdiverse'): bool,
    Optional('--qarecomb'): bool,
    Optional('--qirecomb'): bool,
    Optional('--regressor'): Or(None, str),
    Optional('--rnd_splits'): Use(int),
    Optional('--rnd_state'): Use(int),
    Optional('--save'): Use(int),
    Optional('--sep'): Or(None, str),
    Optional('--show'): Use(int),
    Optional('--sort'): Use(int),
    Optional('--ssm'): bool,
    Optional('--sto'): Or(None, str),
    Optional('--tdiverse'): bool,
    Optional('--temp'): Use(float),
    Optional('--test_size'): Use(float),
    Optional('--threads'): Or(None, Use(int)),
    Optional('--train_size'): Use(float),
    Optional('--trecomb'): bool,
    Optional('--usecsv'): bool,
    Optional('--ts'): Or(None, str),
    Optional('--wt'): Or(None, str),
    Optional('--wt_pos'): Use(int),
    Optional('--y_wt'): Or(None, Use(float)),
    Optional('aaidx'): bool,
    Optional('param_inference'): bool,
    Optional('hybrid'): bool,
    Optional('directevo'): bool,
    Optional('encode'): bool,
    Optional('extrapolation'): bool,
    Optional('low_n'): bool,
    Optional('mklsts'): bool,
    Optional('mkps'): bool,
    Optional('ml'): bool,
    Optional('MODELS'): Or(None, Use(int)),
    Optional('onehot'): bool,
    Optional('reformat_csv'): bool,
    Optional('save_msa_info'): bool,
    Optional('shift_pos'): bool,
    Optional('sto2a2m'): bool,
    Optional('train_and_save'): bool,
})


def validate(args):
    try:
        args = schema.validate(args)
        return args
    except SchemaError as e:
        exit(e)


def run_main():
    """
    Entry point for pip-installed version.
    """
    arguments = docopt(__doc__, version=__version__)
    start_time = time.time()
    logger.debug(f'main.py __name__: {__name__}, version: {__version__}')
    logger.debug(str(argv)[1:-1].replace("\'", "").replace(",", ""))
    logger.debug(f'\n{arguments}')
    arguments = validate(arguments)
    if arguments['directevo']:
        run_pypef_utils(arguments)
    elif arguments['ml']:
        run_pypef_pure_ml(arguments)
    elif arguments['hybrid'] or arguments['param_inference'] or arguments['save_msa_info']:
        run_pypef_hybrid_modeling(arguments)
    else:
        run_pypef_utils(arguments)

    elapsed = str(timedelta(seconds=time.time()-start_time)).split(".")[0]
    elapsed = f'{elapsed.split(":")[0]} h {elapsed.split(":")[1]} min {elapsed.split(":")[2]} s'
    logger.info(f'Done! (Run time: {elapsed})')


if __name__ == '__main__':
    """
    Entry point for direct file run.
    """
    run_main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Modules for performing random evolution walks
similar as presented by Biswas et al.
"""


from __future__ import annotations
import os
import re
import random

import matplotlib.pyplot as plt
import numpy as np
import warnings
from adjustText import adjust_text
import logging
logger = logging.getLogger('pypef.utils.directed_evolution')

from pypef.ml.regression import predict
from pypef.dca.hybrid_model import predict_directed_evolution

# ignoring warnings of scikit-learn regression
warnings.filterwarnings(action='ignore', category=RuntimeWarning, module='sklearn')
warnings.filterwarnings(action='ignore', category=UserWarning, module='sklearn')


class DirectedEvolution:
    # Class attributes (None)
    def __init__(  # Instance attributes
            self,
            ml_or_hybrid: str,
            encoding: str,
            s_wt: str,
            y_wt: float,
            single_vars: list,
            num_iterations: int,
            num_trajectories: int,
            amino_acids: list,
            temp: float,
            path: str,
            model: str = None,
            no_fft: bool = False,
            dca_encoder: str | None = None,
            usecsv: bool = False,
            csvaa: bool = False,
            negative: bool = False
    ):
        """
        Runs in silico directed evolution and plots and writes trajectories.

        Parameters
        ----------
        ml_or_hybrid: str
            'ml' or 'hybrid'
        encoding: str
            'aaidx' or 'dca'
        s_wt: str
            WT sequence, s_wt = get_wt_sequence(arguments['--wt'])
        y_wt: float
            WT fitness, y_wt = arguments['--y_wt']
        single_vars:  list
            single substituted protein variants; used for recombination
            of variants. Obtained from the CSV file with get_variants:
            single_variants, single_values, higher_variants, higher_values = \
                get_variants(df, amino_acids, s_wt)
        num_iterations: int
            Number of tried steps in the evolution process
        num_trajectories: int
            Number of independent evolutionary trajectories
        amino_acids: list
            Usually the 20 standard amino acids
        temp: float
            (Boltzmann) 'Temperature' of the Metropolis-Hastings algorithm for
            accepting new trajectory variants
        path: str
            Just current working directory (os.getcwd())
        model: str
            Loaded Pickle file for regression/hybrid modeling.
        no_fft: bool
            If True, not using FFT for AAindex-based encoding
        dca_encoder = None or PLMC object
            dca_encoder = PLMC(
                  params_file=arguments['--plmc_params'],
                  separator=arguments['--sep']
            )
        usecsv: bool
            Using only CSV variants for recombination (but all 20 amino acids)
        csvaa: bool
            Using only CSV variants for recombination (but all amino acids that
            are present CSV)
        negative: bool
            More negative variants define improved variants
        """
        self.ml_or_hybrid = ml_or_hybrid
        self.encoding = encoding
        self.s_wt = s_wt
        self.y_wt = y_wt
        self.single_vars = single_vars
        self.num_iterations = num_iterations
        self.num_trajectories = num_trajectories
        self.amino_acids = amino_acids
        self.temp = temp
        self.path = path
        self.model = model
        self.no_fft = no_fft  # for AAidx only
        self.dca_encoder = dca_encoder
        self.usecsv = usecsv
        self.csvaa = csvaa
        self.negative = negative
        self.de_step_counter = 0  # DE steps
        self.traj_counter = 0  # Trajectory counter

    def mutate_sequence(
            self,
            seq: str,
            prev_mut_loc: int
    ):
        """
        Parameters
        ----------
        seq: str,
            Initial sequence to be mutated, must not be WT Seq but can
            also itself be already substituted (iterative sequence substitutions)
        prev_mut_loc: int
            Previous position mutated, new position will be randomly chosen within
            a range, by default: new_pos = previous_pos +- 8

        Produces a mutant sequence (integer representation), given an initial sequence
        and the previous position of mutation.

        """
        try:
            os.mkdir('EvoTraj')
        except FileExistsError:
            pass

        var_seq_list = []
        rand_loc = random.randint(prev_mut_loc - 8, prev_mut_loc + 8)  # find random position to mutate
        while (rand_loc <= 0) or (rand_loc >= len(seq)):
            rand_loc = random.randint(prev_mut_loc - 8, prev_mut_loc + 8)
        aa_list = self.amino_acids
        if self.usecsv:     # Only perform directed evolution on positional csv variant data,
            pos_list = []   # else: aa_list = amino_acids
            aa_list = []    # overwrite aa_list = self.amino_acids
            for aa_positions_aa in self.single_vars:  # getting each single variant, e.g. of [['L215F'], ['A217N']]
                for variant in aa_positions_aa:  # just unpacking the variant, e.g. ['L215F'] -> 'L215F'
                    pos_int = int(re.findall(r"\d+", variant)[0])
                    if pos_int not in pos_list:
                        pos_list.append(pos_int)
                    if self.csvaa:
                        new_aa = str(variant[-1:])  # new AA from known variant, e.g. 'F' from 'L215F'
                        if new_aa not in aa_list:
                            aa_list.append(new_aa)
                    else:
                        aa_list = self.amino_acids  # new AA can be any of the 20 standard AA's
            # Select closest position to single AA positions:
            # However, this means that it is more probable that starting with lower substitution
            # positions new substitution positions will likely be shifted towards higher positions.
            # And for higher substitution positions new substitutions will likely be at lower positions.
            absolute_difference_function = lambda list_value: abs(list_value - rand_loc)
            try:
                closest_loc = min(pos_list, key=absolute_difference_function)
            except ValueError:
                raise ValueError("No positions for recombination found. Likely no single "
                                 "substituted variants were found in provided .csv file.")
            rand_loc = closest_loc - 1   # - 1 as position is shifted by one when starting with 0 index
        rand_aa = random.choice(aa_list)  # find random amino acid to mutate to
        seq_list = list(seq)
        seq_list[rand_loc] = rand_aa  # update sequence to have new amino acid at randomly chosen position
        seq_m = ''.join(seq_list)
        var = str(rand_loc + 1) + str(rand_aa)
        var_seq_list.append((var, seq_m))  # list of tuples

        return var_seq_list

    @staticmethod
    def assert_trajectory_sequences(v_traj, s_traj):
        """
        Making sure that sequence mutations have been introduced correctly
        (for last sequence only).
        """
        for i, variant in enumerate(v_traj[1:]):  # [1:] as not checking for WT
            variant_position = int(re.findall(r"\d+", variant)[0]) - 1
            variant_amino_acid = str(variant[-1])
            assert variant_amino_acid == s_traj[i+1][variant_position]  # checking AA of last trajectory sequence

    def in_silico_de(self):
        """
        Perform directed evolution by randomly selecting a sequence
        position for substitution and randomly choose the amino acid
        to substitute to. New sequence gets accepted if meeting the
        Metropolis criterion and will be taken for new substitution
        iteration. Metropolis-Hastings-driven directed evolution,
        similar to Biswas et al.:
        Low-N protein engineering with data-efficient deep learning,
        see https://github.com/ivanjayapurna/low-n-protein-engineering/tree/master/directed-evo
        """
        # iterate through the trial mutation steps for the directed evolution trajectory
        # m = 1 (only 1 mutation per step) instead of (np.random.poisson(2) + 1)
        v_traj, s_traj, y_traj = [], [], []
        v_traj.append('WT')
        y_traj.append(self.y_wt)
        s_traj.append(self.s_wt)
        accepted = 0
        logger.info(f"Step 0: WT --> {self.y_wt:.3f}")
        for iteration in range(self.num_iterations):  # num_iterations
            self.de_step_counter = iteration

            if accepted == 0:
                prior_mutation_location = random.randint(0, len(self.s_wt))  # not really "prior" as first
            else:  # get prior mutation position
                prior_mutation_location = int(re.findall(r"\d+", v_traj[-1])[0])
            prior_y = y_traj[-1]  # prior y, always at [-1]
            prior_sequence = s_traj[-1]  # prior sequence, always at [-1]

            new_var_seq = self.mutate_sequence(
                seq=prior_sequence,
                prev_mut_loc=prior_mutation_location
            )

            new_variant = new_var_seq[0][0]  # int + string char, e.g. '17A'
            new_full_variant = str(self.s_wt[int(new_variant[:-1])-1]) + new_variant  # full variant name, e.g. 'F17A'
            new_sequence = new_var_seq[0][1]
            # encode and predict new sequence fitness
            if self.ml_or_hybrid == 'ml':
                predictions = predict(  # AAidx, OneHot, or DCA-based pure ML prediction
                    path=self.path,
                    model=self.model,
                    encoding=self.encoding,
                    variants=np.atleast_1d(new_full_variant),
                    sequences=np.atleast_1d(new_sequence),
                    no_fft=self.no_fft,
                    couplings_file=self.dca_encoder
                )

            else:  # hybrid modeling and prediction
                predictions = predict_directed_evolution(
                    encoder=self.dca_encoder,
                    variant=self.s_wt[int(new_variant[:-1]) - 1] + new_variant,
                    sequence=new_sequence,
                    hybrid_model_data_pkl=self.model
                )
            if predictions != 'skip':
                logger.info(f"Step {self.de_step_counter + 1}: "
                            f"{self.s_wt[int(new_variant[:-1]) - 1]}{new_variant} --> {predictions[0][0]:.3f}")
            else:  # skip if variant cannot be encoded by DCA-based encoding technique
                logger.info(f"Step {self.de_step_counter + 1}: "
                            f"{self.s_wt[int(new_variant[:-1]) - 1]}{new_variant} --> {predictions}")
                continue
            new_y, new_var = predictions[0][0], predictions[0][1]  # new_var == new_variant nonetheless
            # probability function for trial sequence
            # The lower the fitness (y) of the new variant, the higher are the chances to get excluded
            with warnings.catch_warnings():  # catching Overflow warning
                warnings.simplefilter("ignore")
                try:
                    boltz = np.exp(((new_y - prior_y) / self.temp), dtype=np.longfloat)
                    if self.negative:
                        boltz = np.exp((-(new_y - prior_y) / self.temp), dtype=np.longfloat)
                except OverflowError:
                    boltz = 1
            p = min(1, boltz)
            rand_var = random.random()  # random float between 0 and 1
            if rand_var < p:  # Metropolis-Hastings update selection criterion, else do nothing (do not accept variant)
                v_traj.append(new_var)       # update the variant naming trajectory
                y_traj.append(new_y)         # update the fitness trajectory records
                s_traj.append(new_sequence)  # update the sequence trajectory records
                accepted += 1

        self.assert_trajectory_sequences(v_traj, s_traj)

        return v_traj, s_traj, y_traj

    def run_de_trajectories(self):
        """
        Runs the directed evolution by addressing the in_silico_de
        function and plots the evolution trajectories.
        """
        v_records = []  # initialize list of sequence variant names
        s_records = []  # initialize list of sequence records
        y_records = []  # initialize list of fitness score records
        #   i = counter, iterate through however many mutation trajectories we want to sample
        for i in range(self.num_trajectories):
            self.traj_counter = i
            # call the directed evolution function, outputting the trajectory
            # sequence and fitness score records
            v_traj, s_traj, y_traj = self.in_silico_de()
            v_records.append(v_traj)  # update variant naming full mutagenesis trajectory
            s_records.append(s_traj)  # update the sequence full mutagenesis trajectory
            y_records.append(y_traj)  # update the fitness full mutagenesis trajectory

        return s_records, v_records, y_records

    def plot_trajectories(self):
        """
        Plots evolutionary trajectories and saves steps
        in CSV file.
        """
        s_records, v_records, y_records = self.run_de_trajectories()
        # Idea: Standardizing DCA-HybridModel predictions as just trained by Spearman's rho
        # e.g., meaning that fitness values could differ only at the 6th decimal place and only
        # predicted fitness ranks matter and not associated fitness values
        fig, ax = plt.subplots(figsize=(10,6))  # figsize=(10, 6)
        ax.locator_params(integer=True)
        y_records_ = []
        for i, fitness_array in enumerate(y_records):
            ax.plot(np.arange(1, len(fitness_array) + 1, 1), fitness_array,
                    '-o', alpha=0.7, markeredgecolor='black', label='EvoTraj' + str(i + 1))
            y_records_.append(fitness_array)
        label_x_y_name = []
        traj_max_len = 0
        for i, v_record in enumerate(v_records):  # i = 1, 2, 3, .., ; v_record = variant label array
            for j, v in enumerate(v_record):      # j = 1, 2, 3, ..., ; v = variant name; y_records[i][j] = fitness
                if len(v_record) > traj_max_len:
                    traj_max_len = len(v_record)
                if i == 0:                      # j + 1 -> x-axis position shifted by 1
                    label_x_y_name.append(ax.text(j + 1, y_records_[i][j], v, size=7))
                else:
                    if v != 'WT':  # only plot 'WT' name once at i == 0
                        label_x_y_name.append(ax.text(j + 1, y_records_[i][j], v, size=7))
        adjust_text(label_x_y_name, only_move={'points': 'y', 'text': 'y'}, force_points=0.6)
        ax.legend()
        plt.xticks(np.arange(1,  traj_max_len + 1, 1), np.arange(1, traj_max_len + 1, 1))

        plt.ylabel('Predicted fitness')
        plt.xlabel('Mutation trial steps')
        plt.tight_layout()
        plt.savefig(str(self.model) + '_DE_trajectories.png', dpi=500)
        plt.clf()

        with open(os.path.join('EvoTraj', 'Trajectories.csv'), 'w') as file:
            file.write('Trajectory;Variant;Sequence;Fitness\n')
            for i in range(self.num_trajectories):
                v_records_str = str(v_records[i])[1:-1].replace("'", "")
                s_records_str = str(s_records[i])[1:-1].replace("'", "")
                y_records_str = str(y_records[i])[1:-1]
                file.write(f'{i+1};{v_records_str};{s_records_str};{y_records_str}\n')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

import os
import random
import logging
logger = logging.getLogger('pypef.utils.low_n_mutation_extrapolation')

import pandas as pd
import numpy as np
from scipy import stats
import matplotlib.pyplot as plt
import pickle
from tqdm import tqdm

from pypef.ml.regression import cv_regression_options
from pypef.dca.hybrid_model import DCAHybridModel
from pypef.utils.variant_data import process_df_encoding, get_basename


def get_train_sizes(number_variants) -> np.ndarray:
    """
    Generates a list of train sizes to perform low-n with.
    Returns
    -------
    Numpy array of train sizes up to 80% (i.e. 0.8 * N_variants).
    """
    eighty_percent = int(number_variants * 0.8)
    train_sizes = np.sort(np.concatenate([
        np.arange(15, 50, 5), np.arange(50, 100, 10),
        np.arange(100, 150, 20), [160, 200, 250, 300, eighty_percent],
        np.arange(400, 1100, 100)
    ]))
    idx_max = np.where(train_sizes >= eighty_percent)[0][0] + 1
    return train_sizes[:idx_max]


def plot_low_n(
        train_sizes: list,
        avg_spearmanr: list,
        stddev_spearmanr: list,
        plt_name: str = ''
):
    """
    Plot the performance results of the low N engineering task.
    """
    logger.info('Plotting...')
    plt.plot(train_sizes, avg_spearmanr, 'ko--', linewidth=1, markersize=1.5)
    plt.fill_between(
        np.array(train_sizes),
        np.array(avg_spearmanr) + np.array(stddev_spearmanr),
        np.array(avg_spearmanr) - np.array(stddev_spearmanr),
        alpha=0.5
    )
    plt.ylim(0, max(np.array(avg_spearmanr) * 1.1 + np.array(stddev_spearmanr)))
    plt.xlabel('Train sizes')
    plt.ylabel(r"Spearman's $\rho$")

    plt.savefig(plt_name.split(os.sep)[-1] + '.png', dpi=500)
    plt.clf()


def low_n(
        encoded_csv: str,
        cv_regressor: str = None,
        n_runs: int = 10,
        hybrid_modeling: bool = False,
        train_size_train: float = 0.66
):
    """
    Performs the "low N protein engineering task" learning on distinct
    numbers of encoded_variant_sequences-fitness data to predict the
    left out data (full dataset - train dataset). Maximum sizes of
    learning sets is 0.8 * full dataset (and thus maximal size of test
    set 0.2 * full dataset).
    """
    df = pd.read_csv(encoded_csv, sep=';', comment='#')
    if df.shape[1] == 1:
        df = pd.read_csv(encoded_csv, sep=',', comment='#')
    if df.shape[1] == 1:
        df = pd.read_csv(encoded_csv, sep='\t', comment='#')
    if cv_regressor:
        name = 'ml_' + cv_regressor
        if cv_regressor == 'pls_loocv':
            raise SystemError(
                'PLS LOOCV is not (yet) implemented for the extrapolation task. '
                'Please choose another CV regression option.'
            )
        regressor = cv_regression_options(cv_regressor)
    elif hybrid_modeling:
        name = 'hybrid_ridge'
    n_variants = df.shape[0]
    train_sizes = get_train_sizes(n_variants).tolist()
    variants, x, y = process_df_encoding(df)
    if not x.any():
        raise SystemError("Provided CSV file seems to have no encoding columns "
                          "(required input CSV format: column 1: variant, column 2: "
                          "variant fitness value, column 3 and ongoing columns: "
                          "encoding feature values).")

    avg_spearmanr, stddev_spearmanr = [], []
    # test_sizes = [n_variants - size for size in train_sizes]
    if hybrid_modeling:
        logger.info('Using first CSV row/entry as wild type reference...')
    for size in tqdm(train_sizes):
        spearmanr_nruns = []
        for _ in range(n_runs):
            train_idxs = random.sample(range(n_variants - 1), int(size))
            test_idxs = []
            for n in range(n_variants - 1):
                if n not in train_idxs:
                    test_idxs.append(n)
            x_train, y_train = x[train_idxs], y[train_idxs]
            x_test, y_test = x[test_idxs], y[test_idxs]

            if hybrid_modeling:
                x_wt = x[0]  # WT should be first CSV variant entry
                hybrid_model = DCAHybridModel(
                    x_train=x_train,
                    y_train=y_train,
                    x_test=x_test,  # only used for adjusting +/- sign of y_dca, can also be None
                    y_test=y_test,  # only used for adjusting +/- sign of y_dca, can also be None
                    x_wt=x_wt
                )
                beta_1, beta_2, reg = hybrid_model.settings(
                    x_train, y_train, train_size_fit=train_size_train
                )
                spearmanr_nruns.append(
                    hybrid_model.spearmanr(
                        y_test,
                        hybrid_model.hybrid_prediction(
                            x_test, reg, beta_1, beta_2
                        )
                    )
                )

            else:  # ML
                regressor.fit(x_train, y_train)
                # Best CV params: best_params = regressor.best_params_
                y_pred = regressor.predict(x_test)
                spearmanr_nruns.append(stats.spearmanr(y_test, y_pred)[0])
        avg_spearmanr.append(np.mean(spearmanr_nruns))
        stddev_spearmanr.append(np.std(spearmanr_nruns, ddof=1))

    plot_low_n(
        train_sizes,
        avg_spearmanr,
        stddev_spearmanr,
        'low_N_' + str(encoded_csv).split('.')[0] + '_' + name
    )

    return train_sizes, avg_spearmanr, stddev_spearmanr


def count_mutation_levels_and_get_dfs(df_encoding) -> tuple:
    """
    The input dataframe (from the sequence encoding CSV file) is split
    according to levels of variant substitutions. Substitution seperator
    is '/'.
    """
    single_variants_index, all_higher_variants_index = [], []
    double_i, triple_i, quadruple_i, quintuple_i, sextuple_i, \
    septuple_i, octuple_i, nonuple_i, higher_nine_i = [], [], [], [], [], [], [], [], []
    for i, row in enumerate(df_encoding.iloc[:, 0]):  # iterate over variant column
        if '/' in row:  # TypeError: argument of type 'float' is not iterable if empty columns are (at end of) CSV
            all_higher_variants_index.append(i)
            if row.count('/') == 1:
                double_i.append(i)
            elif row.count('/') == 2:
                triple_i.append(i)
            elif row.count('/') == 3:
                quadruple_i.append(i)
            elif row.count('/') == 4:
                quintuple_i.append(i)
            elif row.count('/') == 5:
                sextuple_i.append(i)
            elif row.count('/') == 6:
                septuple_i.append(i)
            elif row.count('/') == 7:
                octuple_i.append(i)
            elif row.count('/') == 8:
                nonuple_i.append(i)
            elif row.count('/') >= 9:
                higher_nine_i.append(i)
        else:
            single_variants_index.append(i)
    logger.info(f'\nNo. Singles: {len(single_variants_index)}\nNo. All higher: {len(all_higher_variants_index)}\n'
                f'2: {len(double_i)}\n3: {len(triple_i)}\n4: {len(quadruple_i)}\n'
                f'5: {len(quintuple_i)}\n6: {len(sextuple_i)}\n7: {len(septuple_i)}\n'
                f'8: {len(octuple_i)}\n9: {len(nonuple_i)}\n>=10: {len(higher_nine_i)}')
    return (
        df_encoding.iloc[single_variants_index, :],
        df_encoding.iloc[double_i, :],
        df_encoding.iloc[triple_i, :],
        df_encoding.iloc[quadruple_i, :],
        df_encoding.iloc[quintuple_i, :],
        df_encoding.iloc[sextuple_i, :],
        df_encoding.iloc[septuple_i, :],
        df_encoding.iloc[octuple_i, :],
        df_encoding.iloc[nonuple_i, :],
        df_encoding.iloc[higher_nine_i, :],
        df_encoding.iloc[all_higher_variants_index, :],
    )


def performance_mutation_extrapolation(
        encoded_csv: str,
        cv_regressor: str = None,
        train_size: float = 0.66,
        conc: bool = False,
        save_model: bool = True,
        hybrid_modeling: bool = False
) -> dict:
    """
    Train on distinct mutation levels, e.g. only single-substituted samples
    of encoded_variant_sequences-fitness data to predict distinct levels
    of higher substituted variants (i.e. 1->2, 1->3, 1->4 etc.). Also can
    train on concatenated levels of substitution-fitness data using the flag
    --conc, i.e. conc = True (i.e. 1->2, 1+2->3, 1+2+3->4, etc.).
    """
    df = pd.read_csv(encoded_csv, sep=';', comment='#')
    if df.shape[1] == 1:
        df = pd.read_csv(encoded_csv, sep=',', comment='#')
    if df.shape[1] == 1:
        df = pd.read_csv(encoded_csv, sep='\t', comment='#')

    df_mut_lvl = count_mutation_levels_and_get_dfs(df)
    name = ''
    if save_model:
        try:
            os.mkdir('Pickles')
        except FileExistsError:
            pass
    if hybrid_modeling:
        regressor = None
        name = 'hybrid_ridge_' + get_basename(encoded_csv)
    elif cv_regressor:
        name = 'ml_' + cv_regressor + '_' + get_basename(encoded_csv)
        if cv_regressor == 'pls_loocv':
            raise SystemError(
                'PLS LOOCV is not implemented for the extrapolation '
                'task. Please choose another CV regressor.'
            )
        regressor = cv_regression_options(cv_regressor)
        beta_1, beta_2 = None, None
    else:
        regressor = None
        hybrid_model = None
    data = {}
    collected_levels = []
    for i_m, mutation_level_df in enumerate(df_mut_lvl):
        if mutation_level_df.shape[0] != 0:
            collected_levels.append(i_m)
    train_idx_appended = []
    if len(collected_levels) > 1:
        train_idx = collected_levels[0]
        train_df = df_mut_lvl[train_idx]
        train_variants, x_train, y_train = process_df_encoding(train_df)
        all_higher_df = df_mut_lvl[-1]  # only used for adjusting +/- of y_dca
        all_higher_variants, x_all_higher, y_all_higher = process_df_encoding(all_higher_df)
        if hybrid_modeling:
            x_wt = x_train[0]
            hybrid_model = DCAHybridModel(
                x_train=x_train,
                y_train=y_train,
                x_test=x_all_higher,  # only used for adjusting +/- of y_dca, can also be None but
                y_test=y_all_higher,  # higher risk of wrong sign assignment of beta_1 (y_dca)
                x_wt=x_wt
            )
            beta_1, beta_2, reg = hybrid_model.settings(
                x_train, y_train, train_size_fit=train_size)
            pickle.dump(
                {'hybrid_model': hybrid_model, 'beta_1': beta_1, 'beta_2': beta_2,
                 'spearman_rho': float('nan'), 'regressor': reg},
                open(os.path.join('Pickles', 'HYBRID_LVL_1'), 'wb')
            )
        elif cv_regressor:
            logger.info('Fitting regressor on lvl 1 substitution data...')
            regressor.fit(x_train, y_train)
            if save_model:
                logger.info(f'Saving model as Pickle file: ML_LVL_1')
                pickle.dump(regressor, open(os.path.join('Pickles', 'ML_LVL_1'), 'wb'))
        for i, _ in enumerate(tqdm(collected_levels)):
            if i < len(collected_levels) - 1:  # not last i else error, last entry is: lvl 1 --> all higher variants
                test_idx = collected_levels[i + 1]
                test_df = df_mut_lvl[test_idx]
                test_variants, x_test, y_test = process_df_encoding(test_df)
                if not conc:
                    # For training on distinct iterated level i, uncomment lines below:
                    # train_idx = collected_levels[i]
                    # train_df = self.mutation_level_dfs[train_idx]
                    # train_variants, x_train, y_train = self._process_df_encoding(train_df)
                    if hybrid_modeling:
                        data.update({
                            test_idx + 1:
                                {
                                    'hybrid_model': hybrid_model,
                                    'max_train_lvl': train_idx + 1,
                                    'n_y_train': len(y_train),
                                    'test_lvl': test_idx + 1,
                                    'n_y_test': len(y_test),
                                    'spearman_rho': hybrid_model.spearmanr(
                                        y_test,
                                        hybrid_model.hybrid_prediction(
                                            x_test, reg, beta_1, beta_2
                                        )
                                    ),
                                    'beta_1': beta_1,
                                    'beta_2': beta_2,
                                    'regressor': reg
                                }
                        })
                    else:  # ML
                        data.update({
                            test_idx + 1:
                                {
                                    'regressor': regressor,
                                    'max_train_lvl': train_idx + 1,
                                    'n_y_train': len(y_train),
                                    'test_lvl': test_idx + 1,
                                    'n_y_test': len(y_test),
                                    'spearman_rho': stats.spearmanr(
                                        y_test,                    # Call predict on the BaseSearchCV estimator
                                        regressor.predict(x_test)  # with the best found parameters
                                    )[0]
                                }
                        })

                else:  # conc mode, training on mutational levels i: 1, ..., max(i)-1
                    train_idx_appended.append(collected_levels[i])
                    if i < len(collected_levels) - 2:  # -2 as not the last (all_higher)  ## i != 0 and
                        train_df_appended_conc = pd.DataFrame()
                        for idx in train_idx_appended:
                            train_df_appended_conc = pd.concat(
                                [train_df_appended_conc, df_mut_lvl[idx]])
                        train_variants_conc, x_train_conc, y_train_conc = \
                            process_df_encoding(train_df_appended_conc)
                        if hybrid_modeling:  # updating hybrid model params with newly inputted concatenated train data
                            beta_1_conc, beta_2_conc, reg_conc = hybrid_model.settings(
                                x_train_conc,
                                y_train_conc,
                                train_size_fit=train_size
                            )
                            data.update({
                                test_idx + 1:
                                    {
                                        'hybrid_model': hybrid_model,
                                        'max_train_lvl': train_idx_appended[-1] + 1,
                                        'n_y_train': len(y_train_conc),
                                        'test_lvl': test_idx + 1,
                                        'n_y_test': len(y_test),
                                        'spearman_rho': hybrid_model.spearmanr(
                                            y_test,
                                            hybrid_model.hybrid_prediction(
                                                x_test, reg_conc, beta_1_conc, beta_2_conc
                                            )
                                        ),
                                        'beta_1': beta_1_conc,
                                        'beta_2': beta_2_conc,
                                        'regressor': reg_conc
                                    }
                            })
                        else:  # ML updating pureML regression model params with newly inputted concatenated train data
                            # Fitting regressor on concatenated substitution data
                            regressor.fit(x_train_conc, y_train_conc)
                            data.update({
                                test_idx + 1:
                                {
                                    'max_train_lvl': train_idx_appended[-1] + 1,
                                    'n_y_train': len(y_train_conc),
                                    'test_lvl': test_idx + 1,
                                    'n_y_test': len(y_test),
                                    'spearman_rho': stats.spearmanr(
                                        y_test,  # Call predict on the BaseSearchCV estimator
                                        regressor.predict(x_test)  # with the best found parameters
                                        )[0],
                                    'regressor': regressor
                                }
                            })
    plot_extrapolation(data, name, conc)

    return data


def plot_extrapolation(
        extrapolation_data: dict,
        name: str = '',
        conc=False
):
    """
    Plot extrapolation results.
    """
    logger.info('Plotting...')
    test_lvls, spearman_rhos, label_infos = [], [], []
    for test_lvl, result_dict in extrapolation_data.items():
        if result_dict['spearman_rho'] is np.nan:
            continue
        test_lvls.append(test_lvl)
        spearman_rhos.append(result_dict['spearman_rho'])
        label_infos.append(
            r'$\leq$' + str(result_dict['max_train_lvl']) + r'$\rightarrow$' + str(result_dict['test_lvl']) +
            '\n' + str(result_dict['n_y_train']) + r'$\rightarrow$' + str(result_dict['n_y_test'])
        )
    label_infos[0] = 'Lvl: ' + label_infos[0].split('\n')[0] + '\n' + r'$N$: ' + label_infos[0].split('\n')[1]
    if not conc:
        label_infos[-1] = label_infos[-1][6] + r'$\rightarrow$' + '>' + \
                          label_infos[-1][6] + '\n' + label_infos[-1].split('\n')[1]
    plt.plot(test_lvls, spearman_rhos, 'x--', markeredgecolor='k', linewidth=0.7, markersize=4)
    plt.fill_between(
        np.array(test_lvls),
        np.repeat(min(spearman_rhos), len(spearman_rhos)),
        np.array(spearman_rhos),
        alpha=0.3
    )
    if conc:
        name += '_train_concat_lvls'
    else:
        name += '_train_lvl_1'
    plt.xticks(test_lvls, label_infos, fontsize=5)
    plt.ylabel(r"Spearman's $\rho$")
    name = name.split(os.sep)[-1] + '_extrapolation.png'
    plt.savefig(name, dpi=500)
    plt.clf()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Contains Python code used for the approach presented in our 'hybrid modeling' paper
Preprint available at: https://doi.org/10.1101/2022.06.07.495081
Code available at: https://github.com/Protein-Engineering-Framework/Hybrid_Model

The included class 'CouplingsModel' has been taken from the script 'model.py' as part of the
EVmutation module (https://github.com/debbiemarkslab/EVmutation) written by Thomas Hopf in the
labs of Debora Marks and Chris Sander at Harvard Medical School and modified (shortened).
See also: https://doi.org/10.1038/nbt.3769
Hopf, T. A., Ingraham, J. B., Poelwijk, F.J., Schärfe, C.P.I., Springer, M., Sander, C., & Marks, D. S. (2016).
Mutation effects predicted from sequence co-variation. Nature Biotechnology, in press.

References:
[1] Hopf, T. A., Ingraham, J. B., Poelwijk, F.J., Schärfe, C.P.I., Springer, M., Sander, C., & Marks, D. S.
    Mutation effects predicted from sequence co-variation.
    Nature Biotechnology, 35, 2017, 128–135
    https://doi.org/10.1038/nbt.3769
[2] Hopf T. A., Green A. G., Schubert B., et al.
    The EVcouplings Python framework for coevolutionary sequence analysis.
    Bioinformatics 35, 2019, 1582–1584
    https://doi.org/10.1093/bioinformatics/bty862
[3] Ekeberg, M., Lövkvist, C., Lan, Y., Weigt, M., & Aurell, E.
    Improved contact prediction in proteins: Using pseudolikelihoods to infer Potts models.
    Physical Review E, 87(1), 2013, 012707. doi:10.1103/PhysRevE.87.012707
    https://doi.org/10.1103/PhysRevE.87.012707
"""


import os
from collections.abc import Iterable
import logging
logger = logging.getLogger('pypef.dca.encoding')

import numpy as np
import ray
from tqdm import tqdm
import pickle

from pypef.utils.variant_data import amino_acids

_SLICE = np.s_[:]


class InvalidVariantError(Exception):
    """
    Description
    -----------
    Exception raised when entered variant does not follow the required scheme
    (integer enclosed by two one-letter code representations of amino acids).

    Attributes
    ----------
    variant: str
        Variant that causes the error
    message: str
        Explanation of the error
    """

    def __init__(self, variant: str):
        self.variant = variant
        message = "The entered variant '%s' does not follow the required scheme " \
                  "(integer enclosed by two one letter code representations of amino acids). " \
                  "Check separator or variant." % self.variant
        self.message = message
        super().__init__(self.message)


class EffectiveSiteError(Exception):
    """
    Description
    -----------
    Exception raised when requested position is not implemented in the DCA model.

    Attributes
    ----------
    position: int
        Position that causes the error
    variant: str
        Variant including that position
    message: str
        Explanation of the error
    """

    def __init__(self, position: int, variant: str, verbose: bool = True):
        self.position = position
        self.variant = variant
        self.verbose = verbose
        message = f"The position {self.position} of variant '{self.variant}' is " \
                  f"not an effective site in the DCA model and thus cannot be predicted."
        if self.verbose:
            logger.info(message)
            self.message = message
            super().__init__(self.message)


def is_valid_substitution(substitution: str) -> bool:
    """
    Description
    -----------
    A substitution has to follow the scheme:
    First character: (wild-type/substituted) amino acid in one-letter code representation
    Last character: (introduced) amino acid in one-letter code representation
    In between: position (of substitution)

    If the entered substitution does not follow this scheme (integer enclosed by two one
    letter code representations of amino acids) return False, else return True.

    Parameters
    -----------
    substitution : str
        Substitution as string: Integer enclosed by two letters representing
        the wild-type (first) and variant amino acid (last) in one letter code.

    Returns
    -------
    boolian
    """
    if not substitution[0] in amino_acids:  # not accepting format IntegerAA, e.g., 123D
        return False

    if not substitution[-1] in amino_acids:
        return False

    try:
        int(substitution[1:-1])
    except ValueError:
        return False

    return True


def is_valid_variant(variant: str, separator='/') -> bool:
    """
    Description
    -----------
    Gets the single substitutions of the variant and checks if they follow the required scheme.

    If the entered substitution does not follow this scheme (integer enclosed by two one-
    letter code representations of amino acids) return False, else return True.

    Parameters
    ----------
    variant : str
        Joined string of integers enclosed by two letters representing the wild type
        and variant amino acid in the single letter code. -> Check separator
    separator : str
        Character to split the variant to obtain the single substitutions (default=',').

    Returns
    -------
    boolian
    """
    for substitution in variant.split(separator):
        if not is_valid_substitution(substitution):
            return False

    return True


def get_single_substitutions(variant: str, separator='/') -> Iterable:
    """
    Description
    -----------
    Generator that extracts and returns the single substitutions of the entered variant.

    Parameters
    ----------
    See 'is_valid_variant' for an explanation.

    Returns
    -------
    Generator object
    """
    if is_valid_variant(variant, separator):
        for substitution in variant.split(separator):
            yield substitution

    else:
        raise InvalidVariantError(variant)


class CouplingsModel:
    """
    Class to store parameters of pairwise undirected graphical model of sequences
    and compute evolutionary couplings, sequence statistical energies, etc.
    """
    def __init__(
            self,
            filename,
            precision="float32",
            verbose: bool = False,
            **kwargs
    ):
        """
        Initializes the object with raw values read from binary .Jij file

        Parameters
        ----------
        filename : str
            Binary Jij file containing model parameters from plmc software
        alphabet : str
            Symbols corresponding to model states (e.g. "-ACGT").
        precision : {"float32", "float64"}, default: "float32"
            Sets if input file has single (float32) or double precision (float64)
        """
        self.index_map = None
        self._target_seq = None
        self._index_list = None
        self.x_wt = None
        self.verbose = verbose
        try:
            self.__read_plmc_v2(filename, precision)
        except TypeError or FileNotFoundError:
            raise SystemError(
                "Did not find (specified) PLMC parameter file. "
                "The parameter file is required for DCA-based "
                "encoding and can be provided via the flag "
                "--params PLMC_FILE."
            )
        self.alphabet_map = {s: i for i, s in enumerate(self.alphabet)}

        # in non-gap mode, focus sequence is still coded with a gap character,
        # but gap is not part of model alphabet anymore; so if mapping crashes
        # that means there is a non-alphabet character in sequence array
        # and therefore there is no focus sequence.
        try:
            self.target_seq_mapped = np.array([self.alphabet_map[x] for x in self.target_seq])
            self.has_target_seq = (np.sum(self.target_seq_mapped) > 0)
        except KeyError:
            self.target_seq_mapped = np.zeros(shape=np.shape(self.l), dtype=np.int32)
            self.has_target_seq = False

    def __read_plmc_v2(self, filename, precision):
        """
        Read updated Jij file format from plmc.

        Parameters
        ----------
        filename : str
            Binary Jij file containing model parameters
        precision : {"float32", "float64"}
            Sets if input file has single or double precision

        """
        with open(filename, "rb") as f:
            # model length, number of symbols, valid/invalid sequences
            # and iterations
            self.l, self.num_symbols, self.n_valid, self.n_invalid, self.num_iter = (
                np.fromfile(f, "int32", 5)
            )

            # theta, regularization weights, and effective number of samples
            self.theta, self.lambda_h, self.lambda_j, self.lambda_group, self.n_eff = (
                np.fromfile(f, precision, 5)
            )

            # Read alphabet (make sure we get proper unicode rather than byte string)
            self.alphabet = np.fromfile(
                f, "S1", self.num_symbols
            ).astype("U1")

            # weights of individual sequences (after clustering)
            self.weights = np.fromfile(
                f, precision, self.n_valid + self.n_invalid
            )

            # target sequence and index mapping, again ensure unicode
            self._target_seq = np.fromfile(f, "S1", self.l).astype("U1")
            self.index_list = np.fromfile(f, "int32", self.l)

            # Analyzing Positions included in the PLMC file (based on the MSA)
            not_valid, valid = [], []
            for num in range(self.index_list[0], self.index_list[-1] + 1, 1):
                if num not in self.index_list:
                    not_valid.append(num)
                else:
                    valid.append(num)
            self.wt_aa_pos = []
            for aa, pos in zip(self._target_seq, self.index_list):
                self.wt_aa_pos.append(str(aa) + str(pos))
            if self.verbose:
                logger.info(f'Evaluating gap content of PLMC parameter file... '
                            f'First amino acid position used in the MSA (PLMC params file) is '
                            f'{self._target_seq[0]}{self.index_list[0]} and the last position '
                            f'used is {self._target_seq[-1]}{self.index_list[-1]}.')
                if len(not_valid) > 0:
                    logger.info(f'Further, non-included positions are:\n{str(not_valid)[1:-1]}')
                logger.info(f'Summary of all effective positions represented in the MSA '
                            f'based on wild-type sequence ({len(valid)} encoded positions):\n'
                            f'{str([aa_pos for aa_pos in self.wt_aa_pos])[1:-1]}'.replace("'", ""))

            # single site frequencies f_i and fields h_i
            self.f_i, = np.fromfile(
                f, dtype=(precision, (self.l, self.num_symbols)), count=1
            )

            self.h_i, = np.fromfile(
                f, dtype=(precision, (self.l, self.num_symbols)), count=1
            )

            # pair frequencies f_ij and pair couplings J_ij / J_ij
            self.f_ij = np.zeros(
                (self.l, self.l, self.num_symbols, self.num_symbols)
            )

            self.j_ij = np.zeros(
                (self.l, self.l, self.num_symbols, self.num_symbols)
            )

            for i in range(self.l - 1):
                for j in range(i + 1, self.l):
                    self.f_ij[i, j], = np.fromfile(
                        f, dtype=(precision, (self.num_symbols, self.num_symbols)),
                        count=1
                    )
                    self.f_ij[j, i] = self.f_ij[i, j].T

            for i in range(self.l - 1):
                for j in range(i + 1, self.l):
                    self.j_ij[i, j], = np.fromfile(
                        f, dtype=(precision, (self.num_symbols, self.num_symbols)),
                        count=1
                    )

                    self.j_ij[j, i] = self.j_ij[i, j].T

    def get_target_seq_and_index(self):
        """
        Gets and returns the target sequence of encodeable positions as
        well as the index list of encodeable positions that are the
        corresponding amino acid positions of the wild type sequence (1-indexed).

        Returns
        ----------
        self._target_seq: list
            List of single letter strings of the wild-type amino acids
            at the encodeable positions
        self._index_list: list
            List of integers of encodeable amino acid positions
        """
        return self._target_seq, self._index_list

    @property
    def target_seq(self):
        """
        Target/Focus sequence of model used for delta_hamiltonian
        calculations (including single and double mutation matrices)
        """
        return self._target_seq

    @target_seq.setter
    def target_seq(self, sequence):
        """
        Define a new target sequence

        Parameters
        ----------
        sequence : str, or list of chars
            Define a new default sequence for relative Hamiltonian
            calculations (e.g. energy difference relative to wild-type
            sequence).
            Length of sequence must correspond to model length (self.l)
        """
        if len(sequence) != self.l:
            raise ValueError(f"Sequence length inconsistent with model length: {len(sequence)} != {self.l}")

        if isinstance(sequence, str):
            sequence = list(sequence)

        self._target_seq = np.array(sequence)
        self.target_seq_mapped = np.array([self.alphabet_map[x] for x in self.target_seq])
        self.has_target_seq = True

    @property
    def index_list(self):
        """
        Target/Focus sequence of model used for delta_hamiltonian
        calculations (including single and double mutation matrices)
        """
        return self._index_list

    @index_list.setter
    def index_list(self, mapping):
        """
        Define a new number mapping for sequences

        Parameters
        ----------
        mapping: list of int
            Sequence indices of the positions in the model.
            Length of list must correspond to model length (self.l)
        """
        if len(mapping) != self.l:
            raise ValueError(f"Mapping length inconsistent with model length: {len(mapping)} != {self.l}")

        self._index_list = np.array(mapping)
        self.index_map = {b: a for a, b in enumerate(self.index_list)}

    def __map(self, indices, mapping):
        """
        Applies a mapping either to a single index, or to a list of indices

        Parameters
        ----------
        indices : Iterable of items to be mapped, or single item
        mapping: Dictionary containing mapping into new space

        Returns
        -------
        Iterable, or single item
            Items mapped into new space
        """
        if ((isinstance(indices, Iterable) and not isinstance(indices, str)) or
                (isinstance(indices, str) and len(indices) > 1)):
            return np.array([mapping[i] for i in indices])
        else:
            return mapping[indices]

    def __4d_access(self, matrix, i=None, j=None, a_i=None, a_j=None):
        """
        Provides shortcut access to column pair properties
        (e.g. J_ij or f_ij matrices)

        Parameters
        -----------
        i : Iterable(int) or int
            Position(s) on first matrix axis
        j : Iterable(int) or int
            Position(s) on second matrix axis
        a_i : Iterable(str) or str
            Symbols corresponding to first matrix axis
        a_j : Iterable(str) or str
            Symbols corresponding to second matrix axis

        Returns
        -------
        np.array
            4D matrix "matrix" sliced according to values i, j, a_i and a_j
        """
        i = self.__map(i, self.index_map) if i is not None else _SLICE
        j = self.__map(j, self.index_map) if j is not None else _SLICE
        a_i = self.__map(a_i, self.alphabet_map) if a_i is not None else _SLICE
        a_j = self.__map(a_j, self.alphabet_map) if a_j is not None else _SLICE
        return matrix[i, j, a_i, a_j]

    def __2d_access(self, matrix, i=None, a_i=None):
        """
        Provides shortcut access to single-column properties
        (e.g. f_i or h_i matrices)

        Parameters
        -----------
        i : Iterable(int) or int
            Position(s) on first matrix axis
        a_i : Iterable(str) or str
            Symbols corresponding to first matrix axis

        Returns
        -------
        np.array
            2D matrix "matrix" sliced according to values i and a_i
        """
        i = self.__map(i, self.index_map) if i is not None else _SLICE
        a_i = self.__map(a_i, self.alphabet_map) if a_i is not None else _SLICE
        return matrix[i, a_i]

    def get_jij(self, i=None, j=None, a_i=None, a_j=None):
        """
        Quick access to J_ij matrix with automatic index mapping.
        See __4d_access for explanation of parameters.
        """
        return self.__4d_access(self.j_ij, i, j, a_i, a_j)

    def get_hi(self, i=None, a_i=None):
        """
        Quick access to h_i matrix with automatic index mapping.
        See __2d_access for explanation of parameters.
        """
        return self.__2d_access(self.h_i, i, a_i)


class PLMC(CouplingsModel):
    def __init__(
            self,
            params_file: str,
            separator: str = '/',
            verbose: bool = True
    ):
        """
        Class for performing the 'DCA encoding'.

        Attributes
        ----------
        params_file: str
            Binary parameter file outputted by PLMC.
        """
        super().__init__(filename=params_file)  # inherit functions and variables from class CouplingsModel
        self.verbose = verbose
        self.separator = separator
        target_seq, index = self.get_target_seq_and_index()
        self.x_wt = self.collect_encoded_sequences(target_seq[0] + str(index[0]) + target_seq[0])

    def _get_position_internal(self, position: int):
        """
        Description
        -----------
        Returns the "internal position" of an amino acid, e.g., D19V is the desired substitution,
        but the fasta sequence starts from residue 3, i.e., the first two residues are "missing".
        The DCA model will then recognize D19 as D17. In order to avoid wrong assignments,
        it is inevitable to calculate the "internal position" 'i'.

        Parameters
        ----------
        position : int
            Position of interest

        Returns
        -------
        i : int
            "Internal position" that may differ due to different starting residue.
        None
            If the requested position is not an active site.
        """
        offset = 0
        i = position - offset
        if i in self.index_list:
            return i
        else:
            return None

    def sum_ji(self, i: int, a_i: str, sequence: np.ndarray) -> float:
        """
        Description
        -----------
        Calculates the sum of all site-site interaction terms when site 'i' is occupied with amino acid 'a_i'.

        Parameters
        ----------
        i : int
            "Internal position" see '_get_position_internal' for an explanation.
        a_i : str
            Introduced amino acid at 'i' in one-letter code representation.
        sequence: np.ndarray
            Sequence of the variant as numpy array.

        Returns
        -------
        j_i : float
            Sum J(i) of all site-site interaction terms acting on position 'i' when occupied with 'a_i'.
        """
        j_i_sum = 0.0
        for j, a_j in zip(self.index_list, sequence):
            j_i_sum += self.get_jij(i=i, a_i=a_i, j=j, a_j=a_j)

        return j_i_sum

    @staticmethod
    def _unpack_substitution(substitution: str) -> tuple:
        """
        Description
        -----------
        Turns string representation of variant into tuple.

        Parameters
        ----------
        substitution : str
            Substitution as string: Integer enclosed by two letters representing
            the wild-type (first) and variant amino acid (last) in one letter code.

        Returns
        -------
        substitution : tuple
            (wild-type amino acid, position, variant amino acid)
        """
        return substitution[0], int(substitution[1:-1]), substitution[-1]

    def check_substitution_naming_against_wt(self, substitution: str, variant: str):
        """
        Checks whether the amino acid to substitute of the variant matches
        the amino acid of the wild type at this position.
        """
        if substitution[:-1] not in self.wt_aa_pos:
            wild_type_aa, position, a_i = self._unpack_substitution(substitution)
            raise SystemError(
                f"The variant naming scheme is not fitting to the PLMC "
                f"scheme. Substitution {substitution} of variant {variant} has "
                f"the amino acid {wild_type_aa} at position {position}, which "
                f"does not match the wild type sequence used as target for DCA-"
                f"based coupling parameter file generation. See summary of "
                f"(effective) wild-type positions and amino acids above. Please "
                f"check your input variant data or generate a new parameter file "
                f"for encoding."
            )

    def encode_variant(self, variant: str) -> np.ndarray:
        """
        Description
        -----------
        Encodes the variant using its "DCA representation".

        Parameters
        ----------
        variant : str
            Joined string of integers enclosed by two letters representing the wild type
            and variant amino acid in the single letter code. -> Check separator

        Returns
        -------
        x_var : np.ndarray
            Encoded sequence of the variant.
        """
        sequence = self.target_seq.copy()
        for substitution in get_single_substitutions(variant, self.separator):  # e.g. A123C/D234E --> A123C, D234C
            wild_type_aa, position, a_i = self._unpack_substitution(substitution)

            i = self._get_position_internal(position)
            if not i:
                raise EffectiveSiteError(position, variant, self.verbose)

            self.check_substitution_naming_against_wt(substitution, variant)
            i_mapped = self.index_map[i]
            sequence[i_mapped] = a_i

        x_var = np.zeros(sequence.size, dtype=float)
        for idx, (i, a_i) in enumerate(zip(self.index_list, sequence)):
            x_var[idx] = self.get_hi(i, a_i) + 0.5 * self.sum_ji(i, a_i, sequence)

        return x_var

    def collect_encoded_sequences(self, variants: list) -> np.ndarray:
        """
        Description
        -----------
        Collects all encoded sequences based on input variant names.

        Parameters
        ----------
        variants: list
            Variant name, e.g. 'A13E', or 'D127F' (wild-type sequence
            would be defined by substitution to itself, e.g. 'F17F').

        Returns
        ----------
        encoded_sequences: list
            List of all collected encoded sequences (features) for all
            inputted variants.
        non_encoded_variants_list_pos:list
            Internal array/list position for variants that could not be
            encoded due to the underlying MSA (inter-gap threshold for
            computing of local and coupling PLMC parameters). These list
            positions must then be removed for the corresponding fitness
            arrays/lists for training and testing the model.
        """
        encoded_sequences = []
        if len(np.atleast_1d(variants)) == 1:  # do not show progress bar for single variant
            set_silence = True  # thus, also not for directed evolution
        else:
            set_silence = False
        for i, variant in enumerate(tqdm(np.atleast_1d(variants), disable=set_silence)):
            try:
                encoded_sequences.append(self.encode_variant(variant))
            except EffectiveSiteError:
                encoded_sequences.append([None])

        return np.array(encoded_sequences, dtype='object')


"""
Below: Some helper functions to run the PLMC class and get 
the encoded sequences in parallel (threading) using Ray and
to construct a pandas.DataFrame to store the encoded sequences 
(features) and the associated fitness values in a CSV file.
"""


def save_plmc_dca_encoding_model(params_file, substitution_sep='/'):
    """
    Just converts plmc params from raw binary to
    Pickle-saved PLMC class.
    """
    logger.info("Transforming the provided plmc params file "
                "to a PLMC Pickle file (Pickles/PLMC).")
    plmc = PLMC(
        params_file=params_file,
        separator=substitution_sep,
        verbose=False
    )
    try:
        os.mkdir('Pickles')
    except FileExistsError:
        pass
    pickle.dump({
        'model': plmc,
        'model_type': 'PLMCpureDCA',
        'beta_1': None,
        'beta_2': None,
        'spearman_rho': None,
        'regressor': None
    },
        open(f'Pickles/PLMC', 'wb')
    )


def get_encoded_sequence(
        variant: str,
        dca_encode: PLMC
):
    """
    Description
    -----------
    Gets encoded sequence based on input variant name and a preexisting
    PLMC instance.

    Parameters
    ----------
    variant: str
        Variant name, e.g. 'A13E', or 'D127F'. Wild-type sequence
        is defined by substitution to itself, e.g. 'F17F'.
    dca_encode: PLMC class object
        For encoding sequences, see above: class PLMC.
    """
    try:
        encoded_seq = dca_encode.encode_variant(variant)
    except EffectiveSiteError:  # position not included in processed MSA
        return

    return encoded_seq


@ray.remote
def _get_data_parallel(
        variants: list,
        sequences: list,
        fitnesses: list,
        dca_encode: PLMC,
        data: list
) -> list:
    """
    Description
    -----------
    Get the variant name, the associated fitness value, and its ("DCA"-)encoded sequence.

    Parameters
    ----------
    variants : list
        List of strings containing the variants to be encoded.
    fitnesses : list
        List of floats (1d) containing the fitness values associated to the variants.
    dca_encode : object
        Initialized 'PLMC' class object.
    data : manager.list()
        Manager.list() object to store the output of multiple processors.

    Returns
    -------
    data : manager.list()
        Filled list with variant names, fitnesses, and encoded sequence.
    """
    for i, (variant, sequence, fitness) in enumerate(zip(variants, sequences, fitnesses)):
        try:
            data.append([variant, sequence, dca_encode.encode_variant(variant), fitness])
        except EffectiveSiteError:  # do not append non-encoded sequences and
            pass                    # associated fitness values

    return data


def get_dca_data_parallel(
        variants: list,
        sequences: list,
        fitnesses: list,
        dca_encode: PLMC,
        threads: int,
        verbose=True
) -> tuple[list, list, list, list]:
    """
    Description
    -----------
    This function allows to generate the encoded sequences based on the variants
    given in 'csv_file' in a parallel manner.

    Parameters
    ----------
    variants: list (or np.ndarray)
        Variant names.
    sequences: list (or np.ndarray)
        Variant-associated protein sequences.
    fitnesses: list (or np.ndarray)
        Variant-associated fitness values.
    dca_encode : object
        Initialized 'PLMC' class object.
    threads : int
        Number of processes to be used for parallel execution.
        n_cores = 1 defines no threading (not using Ray).
    verbose: bool
        Logging message on/off.

    Returns
    -------
    data: numpy.ndarray
        Filled numpy array including variant names, fitnesses, and encoded sequences.
    non_effective_subs: list
        List of variant names that cannot be used for modelling as they are not effective
        positions in the underlying MSA used for generating local and coupling terms.
    """
    if verbose:
        logger.info(f'{len(variants)} input variants. Encoding variant sequences using parameters '
                    f'taken from plmc generated file. This might take some time...')

    idxs_nan = np.array([i for i, b in enumerate(np.isnan(fitnesses)) if b])  # find fitness NaNs
    if idxs_nan.size > 0:  # remove NaNs if present
        logger.info(f'Fitness NaNs are: {idxs_nan}')
        fitnesses = np.delete(fitnesses, idxs_nan)
        variants = np.delete(variants, idxs_nan)

    variants_split = np.array_split(variants, threads)    # split array in n_cores parts
    sequences_split = np.array_split(sequences, threads)  # for Ray parallelization
    fitnesses_split = np.array_split(fitnesses, threads)
    results = ray.get([
        _get_data_parallel.remote(
            variants_split[i],
            sequences_split[i],
            fitnesses_split[i],
            dca_encode,
            []
        ) for i in range(len(variants_split))
    ])

    data = [item for sublist in results for item in sublist]  # fusing all the individual results
    variants = [item[0] for item in data]
    sequences = [item[1] for item in data]
    xs = [item[2] for item in data]
    fitnesses = [item[3] for item in data]

    if verbose:
        logger.info(f'{len(data)} variants after NaN-valued and non-effective '
                    f'site-substituted variant (EffectiveSiteError) dropping.')

    return variants, sequences, xs, fitnesses



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

import os
from os import listdir
from os.path import isfile, join
import ray
import logging
logger = logging.getLogger('pypef.ml.ml_run')

from pypef.ml.parallelization import aaindex_performance_parallel
from pypef.ml.regression import (
    read_models, formatted_output, performance_list, save_model, predict, predict_ts
)
from pypef.utils.to_file import predictions_out
from pypef.utils.low_n_mutation_extrapolation import low_n, performance_mutation_extrapolation


def run_pypef_pure_ml(arguments):
    """
    Running the program, importing all required self-made modules and
    running them dependent on user-passed input arguments using docopt
    for argument parsing.
    """
    threads = abs(arguments['--threads']) if arguments['--threads'] is not None else 1
    threads = threads + 1 if threads == 0 else threads
    if threads > 1:
        ray.init()
    if arguments['--show']:
        if arguments['MODELS'] != str(5):
            try:
                print(read_models(int(arguments['MODELS'])))
            except ValueError:
                print(read_models(5))
            except TypeError:
                print(read_models(5))
        else:
            print(read_models(5))

    else:
        if arguments['--ls'] is not None and arguments['--ts'] is not None:  # LS --> TS
            if arguments['--model'] is None:
                path = os.getcwd()
                try:
                    t_save = int(arguments['--save'])
                except ValueError:
                    t_save = 5
                # Parallelization of AAindex iteration if threads is not None (but int)
                if threads > 1 and arguments['--encoding'] == 'aaidx':
                    logger.info(f'Using {threads} threads for parallel computing. Running...')
                    encoding_performance_list = aaindex_performance_parallel(
                        train_set=arguments['--ls'],
                        test_set=arguments['--ts'],
                        cores=threads,
                        regressor=arguments['--regressor'],
                        no_fft=arguments['--nofft'],
                        sort=arguments['--sort']
                    )

                else:  # run using a single core or use onehot or DCA-based encoding for modeling
                    encoding_performance_list = performance_list(
                        train_set=arguments['--ls'],
                        test_set=arguments['--ts'],
                        encoding=arguments['--encoding'],
                        regressor=arguments['--regressor'],
                        no_fft=arguments['--nofft'],
                        sort=arguments['--sort'],
                        couplings_file=arguments['--params'],  # only for DCA
                        threads=threads  # only for DCA
                    )

                formatted_output(
                    performance_list=encoding_performance_list,
                    no_fft=arguments['--nofft'],
                    minimum_r2=-1E9
                )

                # save_model encodes variants again (possibly change)
                save_model(
                    path=path,
                    performances=encoding_performance_list,
                    training_set=arguments['--ls'],
                    test_set=arguments['--ts'],
                    threshold=t_save,
                    encoding=arguments['--encoding'],
                    regressor=arguments['--regressor'],
                    no_fft=arguments['--nofft'],
                    train_on_all=arguments['--all'],
                    couplings_file=arguments['--params'],  # only for DCA
                    threads=threads,  # only for DCA
                    label=arguments['--label']
                )

        # Prediction of single .fasta file
        elif arguments['--ts'] is not None and arguments['--model'] is not None:
            predict_ts(
                path=os.getcwd(),
                model=arguments['--model'],
                test_set=arguments['--ts'],
                encoding=arguments['--encoding'],
                no_fft=arguments['--nofft'],
                couplings_file=arguments['--params'],  # only for DCA
                label=arguments['--label'],
                threads=threads
            )

        elif arguments['--ps'] is not None and arguments['--model'] is not None:
            predictions = predict(
                path=os.getcwd(),
                prediction_set=arguments['--ps'],
                model=arguments['--model'],
                encoding=arguments['--encoding'],
                mult_path=None,
                no_fft=arguments['--nofft'],
                couplings_file=arguments['--params'],  # only for DCA
                threads=threads  # only for DCA
            )
            if predictions == 'skip' and not arguments['--params']:
                raise SystemError("No couplings file provided. DCA-based sequence encoding "
                                  "requires a (plmc or GREMLIN) parameter file.")
            if arguments['--negative']:
                predictions = sorted(predictions, key=lambda x: x[0], reverse=False)
            predictions_out(
                predictions=predictions,
                model=arguments['--model'],
                prediction_set=arguments['--ps']
            )

        # Prediction on recombinant/diverse variant folder data
        elif arguments['--pmult'] and arguments['--model'] is not None:
            path = os.getcwd()
            recombs_total = []
            recomb_d, recomb_t, recomb_qa, recomb_qi = \
                '/Recomb_Double_Split/', '/Recomb_Triple_Split/', \
                '/Recomb_Quadruple_Split/', '/Recomb_Quintuple_Split/'
            diverse_d, diverse_t, diverse_q = \
                '/Diverse_Double_Split/', '/Diverse_Triple_Split/', '/Diverse_Quadruple_Split/'
            if arguments['--drecomb']:
                recombs_total.append(recomb_d)
            if arguments['--trecomb']:
                recombs_total.append(recomb_t)
            if arguments['--qarecomb']:
                recombs_total.append(recomb_qa)
            if arguments['--qirecomb']:
                recombs_total.append(recomb_qi)
            if arguments['--ddiverse']:
                recombs_total.append(diverse_d)
            if arguments['--tdiverse']:
                recombs_total.append(diverse_t)
            if arguments['--qdiverse']:
                recombs_total.append(diverse_q)
            if arguments['--drecomb'] is False \
                    and arguments['--trecomb'] is False \
                    and arguments['--qarecomb'] is False \
                    and arguments['--qirecomb'] is False \
                    and arguments['--ddiverse'] is False \
                    and arguments['--tdiverse'] is False \
                    and arguments['--qdiverse'] is False:
                raise SystemError('Define prediction target for --pmult, e.g. --pmult --drecomb.')

            for args in recombs_total:
                predictions_total = []
                logger.info(f'Running predictions for variant-sequence files in directory {args[1:-1]}...')
                path_recomb = path + args
                files = [path_recomb + f for f in listdir(path_recomb)
                         if isfile(join(path_recomb, f)) if f.endswith('.fasta')]
                for i, file in enumerate(files):
                    logger.info(f'Encoding files ({i+1}/{len(files)}) for prediction...')
                    predictions = predict(
                        path=path,
                        prediction_set=file,
                        model=arguments['--model'],
                        encoding=arguments['--encoding'],
                        mult_path=path_recomb,
                        no_fft=arguments['--nofft'],
                        couplings_file=arguments['--params'],  # only for DCA
                        threads=threads  # only for DCA
                    )
                    for pred in predictions:
                        predictions_total.append(pred)  # if array gets too large?
                predictions_total = list(dict.fromkeys(predictions_total))  # removing duplicates from list
                if arguments['--negative']:
                    predictions_total = sorted(predictions_total, key=lambda x: x[0], reverse=False)
                else:
                    predictions_total = sorted(predictions_total, key=lambda x: x[0], reverse=True)

                predictions_out(
                    predictions=predictions_total,
                    model=arguments['--model'],
                    prediction_set=f'Top{args[1:-1]}',
                    path=path_recomb
                )

        elif arguments['extrapolation']:
            performance_mutation_extrapolation(
                encoded_csv=arguments['--input'],
                cv_regressor=arguments['--regressor'],
                conc=arguments['--conc']
            )

        elif arguments['low_n']:
            low_n(
                encoded_csv=arguments['--input'],
                cv_regressor=arguments['--regressor']
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Modules for creating training and test sets
from input CSV file (with value separators sep=',' or sep=';')
having the CSV-format

HEADER_VARIANTS;HEADER_VARIANTS_FITNESS
VARIANT_1;FITNESS_VALUE_1
VARIANT_2;FITNESS_VALUE_2
...

according to the self devised training and test set convention
> VARIANT_NAME_1
; FITNESS_1
VARIANT_SEQUENCE_1
> VARIANT_NAME_2
; FITNESS_2
VARIANT_SEQUENCE_2
...
"""

import logging
logger = logging.getLogger('pypef.utils.learning_test_sets')

import numpy as np
import random
import pandas as pd
import re


def csv_input(csv_file):
    """
    Gets input data from defined .csv file (that contains variant names and fitness labels)
    """
    if csv_file is None:
        raise FileNotFoundError(
            f'Did not find (specified) csv file! '
            f'Used csv input file instead: {str(csv_file)}.'
        )
    return csv_file


def drop_rows(
        csv_file,
        amino_acids,
        threshold_drop,
        csv_sep: str = ';',
        mutation_sep: str = '/'
):
    """
    Drops rows from .csv data if below defined fitness threshold or if
    amino acid/variant name is unknown or if fitness label is not a digit.
    """
    separator = ';'
    try:
        df_raw = pd.read_csv(csv_file, sep=separator, usecols=[0, 1])
    except ValueError:
        separator = ','
        df_raw = pd.read_csv(csv_file, sep=separator, usecols=[0, 1])
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Specify the input CSV file containing the variant-fitness data. "
            f"Required CSV format: variant{csv_sep}fitness.")

    label = df_raw.iloc[:, 1]
    sequence = df_raw.iloc[:, 0]

    dropping_rows = []

    for i, row in enumerate(label):
        try:
            row = float(row)
            if row < threshold_drop:
                dropping_rows.append(i)
        except ValueError:
            dropping_rows.append(i)

    for i, variant in enumerate(sequence):
        try:
            if mutation_sep in variant:
                m = re.split(rf'{mutation_sep}', variant)
                for a, splits in enumerate(m):
                    if splits[0].isdigit() and variant[-1] in amino_acids:
                        continue
                    elif splits[0] not in amino_acids or splits[-1] not in amino_acids:
                        if i not in dropping_rows:
                            dropping_rows.append(i)
            else:
                if ',' in variant or ';' in variant or '\t' in variant:
                    raise SystemError("Found invalid characters (';', ',', or tabulator) in variants. "
                                      "Check the --mutation_sep flag and try specifying it, e.g. --mutation_sep \',\'.")
                if variant[0].isdigit() and variant[-1] in amino_acids:
                    continue
                elif variant not in ['wt', 'wild_type']:
                    if variant[0] not in amino_acids or variant[-1] not in amino_acids:
                        dropping_rows.append(i)

        except TypeError:
            raise TypeError('You might consider checking the input .csv for empty first two columns,'
                            ' e.g. in the last row.')

    logger.info(f'No. of dropped rows: {len(dropping_rows)}. '
                f'Total given variants (if provided plus WT): {len(df_raw)}')

    df = df_raw.drop(dropping_rows)
    df.dropna(inplace=True)
    df.reset_index(drop=True, inplace=True)

    return df


def get_variants(
        df,
        amino_acids,
        wild_type_sequence,
        mutation_sep: str = '/'
):
    """
    Gets variants and divides and counts the variant data for single substituted
    and higher substituted variants. Raises NameError if variant naming is not
    matching the given wild-type sequence, e.g. if variant A17C would define
    a substitution at residue Ala-17 to Cys but the wild-type sequence has no Ala
    at position 17.
    """
    x = df.iloc[:, 0]
    y = df.iloc[:, 1]
    wt_position = None
    single_variants, higher_variants, index_higher, index_lower, \
        higher_values, single_values = [], [], [], [], [], []
    single, double, triple, quadruple, quintuple, sextuple, septuple,\
        octuple, nonuple, decuple, higher = 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    for i, variant in enumerate(x):
        if mutation_sep in variant:
            count = variant.count(mutation_sep)
            if count == 1:
                double += 1
            elif count == 2:
                triple += 1
            elif count == 3:
                quadruple += 1
            elif count == 4:
                quintuple += 1
            elif count == 5:
                sextuple += 1
            elif count == 6:
                septuple += 1
            elif count == 7:
                octuple += 1
            elif count == 8:
                nonuple += 1
            elif count == 9:
                decuple += 1
            else:
                higher += 1
            m = re.split(rf'{mutation_sep}', variant)
            for a, splits in enumerate(m):
                if splits[0].isdigit() or splits[0] in amino_acids and splits[-1] in amino_acids:
                    new = int(re.findall(r'\d+', splits)[0])
                    if splits[0] in amino_acids:
                        if splits[0] != wild_type_sequence[new - 1]:
                            raise NameError(
                                'Position of amino acids in given sequence does not match the given '
                                'positions in the input data! E.g. see position {} and position {} being {} '
                                'in the given sequence'.format(variant, new, wild_type_sequence[new - 1])
                            )
                    higher_var = wild_type_sequence[new - 1] + str(new) + str(splits[-1])
                    m[a] = higher_var
                    if a == len(m) - 1:
                        higher_variants.append(m)
                        if i not in index_higher:
                            index_higher.append(i)
        else:
            single += 1
            if variant.upper() == 'WT' or variant.upper() == 'WILD_TYPE':
                wt_position = i

                continue

            elif variant[0].isdigit() or variant[0] in amino_acids and variant[-1] in amino_acids:
                try:
                    num = int(re.findall(r'\d+', variant)[0])
                except IndexError:
                    raise IndexError('Wrong input format. Please check if the input CSV corresponds to the '
                                     'required input style (while the wild-type protein must be designated as \'WT\').')
                if variant[0] in amino_acids:
                    try:
                        if variant[0] != wild_type_sequence[num - 1]:
                            raise NameError('Position of amino acids in given sequence does not match the given '
                                            'positions in the input data! E.g. see position {} and position {} being {}'
                                            ' in the given sequence.'.format(variant, num, wild_type_sequence[num - 1]))
                    except IndexError:
                        raise IndexError("Found variant sequence position {} in data which "
                                         "is out of range of wild-type sequence length.".format(str(num)))
                try:
                    full_variant = wild_type_sequence[num - 1] + str(num) + str(variant[-1])
                except IndexError:
                    raise IndexError("Found variant sequence position {} in data which "
                                     "is out of range of wild-type sequence length.".format(str(num)))
                single_variants.append([full_variant])
                if i not in index_lower:
                    index_lower.append(i)
    logger.info(
        '\nSingle (for mklsts if provided plus WT): {}\nDouble: {}\nTriple: {}\nQuadruple: {}\nQuintuple: {}\n'
        'Sextuple: {}\nSeptuple: {}\nOctuple: {}\nNonuple: {}\nDecuple: {}\nHigher (>Decuple): {}'.format(
            single, double, triple, quadruple, quintuple, sextuple, septuple, octuple, nonuple, decuple, higher
        )
    )
    for vals in y[index_higher]:
        higher_values.append(vals)
    for vals in y[index_lower]:
        single_values.append(vals)
    if wt_position is not None:
        single_variants.append(['WT'])
        single_values.append(y[wt_position])

    single_variants, single_values = tuple(single_variants), tuple(single_values)
    higher_variants, higher_values = tuple(higher_variants), tuple(higher_values)

    return single_variants, single_values, higher_variants, higher_values


def make_sub_ls_ts(
        single_variants,
        single_values,
        higher_variants,
        higher_values,
        directed_evolution=False):
    """
    Creates learning and test sets, fills learning set with single substituted variants and splits
    rest (higher substituted) for learning and test sets: 3/4 to LS and 1/4 to TS
    """
    logger.info(f'No. of single substituted variants (if provided plus WT): {len(single_variants)}.'
                f'No. of values: {len(single_values)}.')
    logger.info(f'No. of higher substituted variants: {len(higher_variants)}. '
                f'No. of values: {len(higher_values)}.')

    if len(single_values) != len(single_variants):
        logger.info(f'Error due to different lengths for given variants and label! '
                    'No. of single substituted variants: {len(single_variants)}. '
                    'Number of given values: {len(single_values)}.')

    if len(higher_values) != len(higher_variants):
        logger.info(f'Error due to different lengths for given variants and label! '
                    f'No. of higher subst. variants: {len(higher_variants)}. '
                    f'Number of given values: {len(higher_values)}.')

    # 1. CREATION OF LS AND TS SPLIT FOR SINGLE FOR LS AND HIGHER VARIANTS FOR TS
    all_variants = single_variants + higher_variants
    all_values = single_values + higher_values
    sub_ts = []  # Substitutions of TS
    values_ts = []  # Values of TS
    sub_ls = []
    values_ls = []

    if directed_evolution is False:
        if len(higher_variants) != 0:
            sub_ls = list(single_variants)  # Substitutions of LS
            values_ls = list(single_values)  # Values of LS
            for i in range(len(higher_variants)):
                if len(higher_variants) < 6:  # if less than 6 higher variants --> all higher variants to TS
                    sub_ts.append(higher_variants[i])
                    values_ts.append(higher_values[i])
                elif (i % 3) == 0 and i != 0:  # 1/4 of higher variants to TS, 3/4 to LS
                    sub_ts.append(higher_variants[i])
                    values_ts.append(higher_values[i])
                else:                       # 3/4 to LS
                    sub_ls.append(higher_variants[i])
                    values_ls.append(higher_values[i])
        else:  # if no higher substituted variants are available split 80%/20%
            random_nums = []
            range_list = np.arange(0, len(all_variants))
            while len(sub_ls) < len(all_variants) * 4 // 5:  # 80 % Learning Set
                random_num = random.choice(range_list)
                if random_num not in random_nums:
                    random_nums.append(random_num)
                    sub_ls.append(all_variants[random_num])
                    values_ls.append(all_values[random_num])
            else:  # 20 % Test Set
                for num in range_list:
                    if num not in random_nums:
                        sub_ts.append(all_variants[num])
                        values_ts.append(all_values[num])

    return sub_ls, values_ls, sub_ts, values_ts


def make_sub_ls_ts_randomly(
        single_variants,
        single_values,
        higher_variants,
        higher_values
):
    """
    Creation of learning set and test set by randomly splitting sets
    """
    length = len(single_variants) + len(higher_variants)
    range_list = np.arange(0, length)

    ts = []
    ls = []
    while len(ls) < length * 4 // 5:
        random_num = random.choice(range_list)
        if random_num not in ls:
            ls.append(random_num)

    for j in range_list:
        if j not in ls:
            ts.append(j)

    combined = single_variants + higher_variants  # substitutions
    combined2 = single_values + higher_values  # values

    sub_ls = []
    values_ls = []
    tot_sub_ls, tot_values_ls = [], []
    tot_sub_ts, tot_values_ts = [], []

    for i in ls:
        sub_ls.append(combined[i])
        values_ls.append(combined2[i])

    sub_ts = []
    values_ts = []
    for j in ts:
        sub_ts.append(combined[j])
        values_ts.append(combined2[j])

    for subs in sub_ls:
        for subs2 in sub_ts:
            if subs == subs2:
                logger.warning(f'\n<Warning> LS and TS overlap for: {subs} - '
                               f'You might want to consider checking the provided '
                               f'datasets for multiple entries')

    tot_sub_ls.append(sub_ls)
    tot_values_ls.append(values_ls)
    tot_sub_ts.append(sub_ts)
    tot_values_ts.append(values_ts)

    return tot_sub_ls[0], tot_values_ls[0], tot_sub_ts[0], tot_values_ts[0]


def make_fasta_ls_ts(
        filename,
        wt_seq,
        substitutions,
        fitness_values
):
    """
    Creates learning and test sets (.fasta style-like files with fitness values
    indicated by starting semicolon ';')

    filename: str
        String for defining the filename for the learning and test set "fasta-like" files.
    wt: str
        Wild-type sequence as string
    substitutions: list
        List of substitutions of a single variant of the format:
            - Single substitution variant, e.g. variant A123C: ['A123C']
            - Higher variants, e.g. variant A123C/D234E/F345G: ['A123C', 'D234E, 'F345G']
            --> Full substitutions list, e.g.: [['A123C'], ['A123C', 'D234E, 'F345G']]
    fitness_values: list
        List of ints/floats of the variant fitness values, e.g. for two variants: [1.4, 0.8]
    """
    myfile = open(filename, 'w')
    for i, var in enumerate(substitutions):  # var are lists of (single or multiple) substitutions
        temp = list(wt_seq)
        name = ''
        separation = 0
        if var == ['WT']:
            name = 'WT'
        else:
            for single_var in var:  # single entries of substitution list
                position_index = int(str(single_var)[1:-1]) - 1
                new_amino_acid = str(single_var)[-1]
                temp[position_index] = new_amino_acid
                # checking if multiple entries are inside list
                if separation == 0:
                    name += single_var
                else:
                    name += '/' + single_var
                separation += 1
        print(f'>{name}', file=myfile)
        print(f';{fitness_values[i]}', file=myfile)
        print(''.join(temp), file=myfile)
        # print(name+';'+str(val[i])+';'+''.join(temp), file=myfile)  # uncomment output: CSV format
    myfile.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution


__version__ = '0.3.1-alpha'
